package AST;

import java.lang.*;
import java.util.*;
import GreenCompiler.*;
import CompilerError.*;
import Interface.*;

public class ArrayType extends Type {

    public ArrayType( Type type, int numDimensions ) {
         super();
         this.type = type;
         this.numDimensions = numDimensions;
         String name = "array(" + type.getName() + ")";
         for (int i = 0; i < numDimensions; i++)
           name = name + "[]";
         setName(name);
         setRealJavaName();
    }

    public String getInterfaceName() {
         return getJavaName();
    }

    public boolean isArrayType() {
         return true;
    }

    public boolean isSupertypeOf(Type other) {
        return getJavaName().compareTo(other.getJavaName()) == 0 ||
               other.getName().compareTo("Nil") == 0;
    }

    public void reify( TypeInterfaceList til, ErrorMsg em ) {
         super.reify(til, em);
         type.reify(til, em);
    }

    public void setRealJavaName() {
        super.setJavaName( NameServer.buildArrayName(type, numDimensions) );
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    public Type getType() {
       return type;
    }

    public int getNumDimensions() {
       return numDimensions;
    }

    private Type type;
    private int numDimensions;

}
