package AST;

import java.lang.*;
import java.io.*;
import java.util.*;
import GreenCompiler.*;
import CompilerError.*;
import Interface.*;
import CompilerError.*;

public class ClassObject {

    public ClassObject(ClassObjectTypeInterface coti) {
        classObjectTypeInterface = coti;
        javaImportList = new Vector();
        toAddString = new StringBuffer();
        
        publicMethodList = new MethodList();
        privateMethodList = new MethodList();
        varList = new ClassObjectVariableList();
        publicConstList = new ConstList();
        privateConstList = new ConstList();
        
    }

    public void addJavaCode(String toAdd) {
        toAddString.append(toAdd);
    }

    public void genInterfaces( String directory, ErrorMsg errorMsg ) {
         Method m;
         Enumeration e = publicMethodList.elements();
         while ( e.hasMoreElements() ) {
            m = (Method ) e.nextElement();
            if ( ! m.getIsRedefinition() ) {
               String outName = directory + m.getMethodInterfaceName() + ".java";
               try {
                  File f = new File(outName);
                  if ( ! f.exists() ) {
                      PrintWriter out = new PrintWriter( 
                          new FileOutputStream(outName) );
                      m.genInterface(out, "co " + classObjectTypeInterface.getName() );
                     out.close();
                  }
               }
               catch( Exception excep ) {
                  errorMsg.show("Error writing to file " + outName );
               }
            }
         }
    }

    public void setAssociatedClass(NormalClass aClass) {
        associatedClass = aClass;
    }

    public void addJavaImport(String importString) {
        javaImportList.addElement(importString);
    }

    public void gen( PrintWriter out, CompilerOptions compilerOptions ) {
        Context context = new Context(null, null, compilerOptions);
        ws.set(out);
        if ( javaImportList.size() > 0 ) {
           Enumeration g = javaImportList.elements();
           while ( g.hasMoreElements() ) 
              out.println( (String ) g.nextElement() );
           out.println();
        }
        out.println("class " + classObjectTypeInterface.getJavaName() + " extends " +
             Type.anyClassObjectTypeInterface.getJavaName() );
        ws.add();
        ws.println("implements");
        ws.add();
        
        classObjectTypeInterface.genInterfaceNameList(out);
        /*
        int size = publicMethodList.getSize();
        Enumeration e = publicMethodList.elements();
        while ( e.hasMoreElements() ) {
           ws.print( ((Method ) e.nextElement()).getMethodInterfaceName() );
           if ( --size > 0 ) {
              out.println(",");
           }
        }
        */
        
        out.println("  {");
        ws.sub();
             
        out.println();
        ws.println("public " + classObjectTypeInterface.getJavaName() + "() {");
        ws.add();
        if ( initMethod != null ) 
           ws.println("_init();");
        ws.sub();
        ws.println("}");
        
        if ( compilerOptions.getUseIRL() ) {
           ws.println("public static _ClassInfo $i = null;");
           ws.println("public static _ClassInfo associateClassInfo;");
           ws.println("public static _ObjectInitMethodInfo initMethodInfo;");
           ws.println("public static _ClassObjectInfo  classObjectInfo;");
           ws.println("public static " + classObjectTypeInterface.getJavaName() + " selfObject;");
       
           ws.println("public static _ObjectMethodInfo ");
           ws.add();
           publicMethodList.genObjectMethodInfoVarList(out, 
                privateMethodList.getSize() > 0 );
           privateMethodList.genObjectMethodInfoVarList(out, false);
           ws.println(";");
           ws.sub();
           
           out.println();
           ws.println("public _Any $callMethod( int methodNumber, $Array$1$_Any args ) {");
           ws.add();
           ws.println("_Any any;");
           out.println();
           ws.println("switch (methodNumber) {");
           ws.add();
           publicMethodList.genCaseCallMethod(out);
           privateMethodList.genCaseCallMethod(out);
           ws.println("default :");
           ws.add();
           ws.println("throw new PackedException( _InternalErrorException.$classObject._new$String(\"Error in $callMethod in a class object\") );");
           ws.sub();
           ws.sub();
           ws.println("}");
           ws.println("return any;");
           ws.sub();
           ws.println("}");
           out.println();
           ws.println("void $set( int ivNumber, _Any value ) {");
           ws.add();
           ws.println("switch ( ivNumber ) {");
           ws.add();
           varList.genCaseSet(out);
           ws.println("default :");
           ws.add();
           ws.println("throw new PackedException( _InternalErrorException.$classObject._new$String(\"Error in $set in a class object\") );");
           ws.sub();
           
           ws.sub();
           ws.println("}");
           ws.sub();
           ws.println("}");
           
           
           out.println();
           ws.println("_Any $get( int ivNumber ) {");
           ws.add();
           ws.println("switch ( ivNumber ) {");
           ws.add();
           varList.genCaseGet(out);
           ws.println("default :");
           ws.add();
           ws.println("throw new PackedException( _InternalErrorException.$classObject._new$String(\"Error in $set in a class object\") );");
           ws.sub();
           ws.sub();
           ws.println("}");
           
           ws.sub();
           ws.println("}");
           
           out.println();
           
           ws.println("static void $init( _ClassInfo p_associateClassInfo, " + 
                classObjectTypeInterface.getJavaName() + " classObject ) {");
           ws.add();
           ws.println("associateClassInfo = p_associateClassInfo;");
           ws.println("selfObject = classObject;");
           ws.println("initMethodInfo = (_ObjectInitMethodInfo ) _ObjectInitMethodInfo.$classObject._new$_MethodBodyInfo$$Array$1$_CodeAnnotation( null, null );");
           
           publicMethodList.genObjectMethodInfo(out);
           privateMethodList.genObjectMethodInfo(out);
           
           ws.println("classObjectInfo = (_ClassObjectInfo ) _ClassObjectInfo.$classObject._new$_Any$_ObjectInitMethodInfo$$Array$1$_ConstantInfo$$Array$1$_ObjectInstanceVariableInfo$$Array$1$_ObjectMethodInfo$$Array$1$_ObjectMethodInfo$_ClassInfo$$Array$1$_CodeAnnotation(");
           ws.add();
           ws.println("selfObject,");
           ws.println("initMethodInfo,");
           
           ws.println("new " + 
                NameServer.buildArrayName("_ConstantInfo", 1) + "(");
           ws.add();
           ws.println("new _ConstantInfo[] {");
           ws.add();
           publicConstList.genConstantInfo(out, privateConstList.getSize() > 0 );
           privateConstList.genConstantInfo(out, false);
           ws.sub();
           ws.println("}");
           ws.sub();
           ws.println("),");
           
           ws.println("new " + 
                NameServer.buildArrayName("_ObjectInstanceVariableInfo", 1) + "(");
           ws.add();
           ws.println("new _ObjectInstanceVariableInfo[] {");
           ws.add();
           varList.genObjectInstanceVariableInfo(out);
           ws.println("}");
           ws.sub();
           ws.println("),");
           ws.sub();
           
           ws.println("new " + 
                NameServer.buildArrayName("_ObjectMethodInfo", 1) + "(");
           ws.add();
           ws.println("new _ObjectMethodInfo[] {");
           ws.add();
           publicMethodList.genObjectMethodInfoList(out, privateMethodList.getSize() > 0);
           privateMethodList.genObjectMethodInfoList(out, false);
           ws.println("}");
           ws.sub();
           ws.println("),");
           ws.sub();
           ws.println("new " + 
                NameServer.buildArrayName("_ObjectMethodInfo", 1) + "(");
           ws.add();
           ws.println("new _ObjectMethodInfo[] {");
           ws.add();
           publicMethodList.genNewObjectMethodInfoList(out);
           ws.println("}");
           ws.sub();
           ws.println("),");
           ws.sub();
           
           //changed
           ws.println( classObjectTypeInterface.getAccessToClassInfo() + "," );
           ws.println("null");
           ws.println(");");
           
           ws.sub();
           
           ws.sub();
           ws.println("}");
           
        }
        ws.sub();
             
        if ( initMethod != null ) {
           out.println();
           initMethod.gen(out, false, context);
        }
        publicMethodList.gen(out, false, context);
        publicConstList.gen(out, context, "public ");
        privateMethodList.gen(out, false, context);
        privateConstList.gen(out, context, "private ");
        varList.gen(out, context);
        out.println(toAddString);
        
         if ( classObjectTypeInterface.getHasThrowMethod() ) {
            ws.add();
            ws.println("public void select( Object e ) {");
            ws.add();
            ws.println("while ( e != null && e instanceof PackedException )");
            ws.add();
            ws.println("e = ((PackedException ) e).getException();");
            ws.sub();
            ParameterList pl = new ParameterList();
            Parameter p = new Parameter("excep");
            p.setType(Type.anyType);
            pl.add(p);
            int n = 0;
            /*
            Hashtable ht = classTypeInterface.multiplePublicSearch(
              NameServer.buildAnySearchName("throw", pl)
              );
            Vector w = new Vector();
            Enumeration f = ht.elements();
            while ( f.hasMoreElements() ) 
               w.addElement(f.nextElement());
            MethodInterfaceList.sortByMethodNumber(w);
            */
            
            Vector w = new Vector();
            Enumeration f = classObjectTypeInterface.getPublicMethodList().anyElements();
            while ( f.hasMoreElements() ) {
               MethodInterface mi = (MethodInterface ) f.nextElement();
               if ( mi.getIsThrow() ) 
                   w.addElement(mi);
            }
               
            f = w.elements();
            while ( f.hasMoreElements() ) {
               MethodInterface mi = (MethodInterface ) f.nextElement();
               Enumeration g = mi.getParamList().elements();
               p = (Parameter ) g.nextElement();
               Type t = p.getType();
               ws.print("");
               if ( n > 0 ) 
                 out.print("else ");
               out.print("if ( e instanceof " + t.getJavaName() );
               out.println(" ) { ");
               ws.add();
               ws.println("_set$_Exception$_AnyClassObject( (_Any ) e, " + 
                  NameServer.getAccessToClassObject( t.getName() ) + ");");
               ws.println( mi.getJavaName() + "( (" + t.getJavaName() +
                  " ) e );");
               ws.println("}");
               ws.sub();
               
               n++;
            }
            ws.println("else {");
            ws.add();
            ws.println( NameServer.getAccessToClassObject("Runtime") +
               "._$removeCatchObjectStack();");
            ws.println("throw new PackedException(e);");
            ws.sub();
            ws.println("}");
            ws.sub();
            ws.println("}");
            ws.sub();
         }
        
        
        out.println("}");
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    public ClassObjectTypeInterface getClassObjectTypeInterface() {
       return classObjectTypeInterface;
    }

    public NormalClass getAssociatedClass() {
       return associatedClass;
    }

    public InitMethod getInitMethod() {
       return initMethod;
    }

    public void setInitMethod( InitMethod initMethod ) {
       this.initMethod = initMethod;
    }

    public MethodList getPublicMethodList() {
       return publicMethodList;
    }

    public void setPublicMethodList( MethodList publicMethodList ) {
       this.publicMethodList = publicMethodList;
    }

    public MethodList getPrivateMethodList() {
       return privateMethodList;
    }

    public void setPrivateMethodList( MethodList privateMethodList ) {
       this.privateMethodList = privateMethodList;
    }

    public ClassObjectVariableList getVarList() {
       return varList;
    }

    public void setVarList( ClassObjectVariableList varList ) {
       this.varList = varList;
    }

    public ConstList getPublicConstList() {
       return publicConstList;
    }

    public void setPublicConstList( ConstList publicConstList ) {
       this.publicConstList = publicConstList;
    }

    public ConstList getPrivateConstList() {
       return privateConstList;
    }

    public void setPrivateConstList( ConstList privateConstList ) {
       this.privateConstList = privateConstList;
    }

    private ClassObjectTypeInterface classObjectTypeInterface;
    private Vector javaImportList;
    private NormalClass associatedClass;
    private InitMethod initMethod;
    private MethodList publicMethodList;
    private MethodList privateMethodList;
    private ClassObjectVariableList varList;
    private ConstList publicConstList;
    private ConstList privateConstList;
    private StringBuffer toAddString;

}
