package AST;

import java.lang.*;
import java.util.*;
import GreenCompiler.*;
import CompilerError.*;
import Interface.*;

public class ClassType extends Type {

    public ClassType( String name, TypeList typeList ) {
       super(name);
       this.typeList = typeList;
       setIsClassObjectType(true);
       setRealJavaName();
    }

    public String getGreenName() {
        if ( typeList == null ) 
           return getName();
        else {
           String s = getName() + "(";
           int size = typeList.getSize();
           Enumeration e = typeList.elements();
           while ( e.hasMoreElements() ) {
              String paramName = ((Type ) e.nextElement()).getGreenName();
              s = s + paramName;
              if ( --size > 0 )
                 s = s + ", ";
           }
           s = s + ")";
           return s;
        }
    }

    public String getInterfaceName() {
        if ( typeList == null ) 
           return getName();
        else 
           return javaName;
    }

    public boolean isSuperclassOf(Type otherType) {
        TypeInterface other = otherType.getTypeInterface();
        if ( ! (other instanceof ClassTypeInterface) ) 
           return false;
        else {
           ClassTypeInterface sub = (ClassTypeInterface ) other;
           while ( sub != null ) 
              if ( sub.getName().compareTo(getName()) == 0 )
                 return true;
              else
                 sub = sub.getSuperclass();
           return false;
       }
    }

    public void setRealJavaName() {
       if ( typeList == null ) 
          super.setJavaName();
       else 
          javaName = NameServer.buildClassName(getName(), typeList);
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    public TypeList getTypeList() {
       return typeList;
    }

    public GenericClass getReifiedClass() {
       return reifiedClass;
    }

    private TypeList typeList;
    private GenericClass reifiedClass;

}
