package AST;

import java.io.*;
import java.util.*;
import GreenCompiler.*;
import CompilerError.*;
import Interface.*;

public class TryStatement extends Statement {

    public TryStatement(Expr expr, StatementList statementList, StatementList finallyBlock) {
       this.expr = expr;
       this.statementList = statementList;
       this.finallyBlock = finallyBlock;
    }

    public void gen( PrintWriter out, Context context ) {
       
       String catchVariableName = context.getCatchVariableName();
       ws.print( expr.getType().getJavaName() + "  " + 
          catchVariableName + " = (" + expr.getType().getJavaName() + " ) ");
       expr.gen(out, context);
       out.println(";");
       
       ws.println("try {");
       ws.add();
       
       String runtimeString = NameServer.getAccessToClassObject("Runtime");
       ws.print( runtimeString +
          "._$addCatchObjectStack$_Catch(" + catchVariableName );
          
       out.println(");");
       /*
       String runtimeString = NameServer.getAccessToClassObject("Runtime");
       ws.print( runtimeString  + "._$setMethodThrownException$_ObjectMethodInfo(");
       out.println( NameServer.getObjectMethodInfoNameVar( context.getCurrentMethod().getName(),
            context.getCurrentMethod().getParamList() ) + ");" );
       */
       
       statementList.gen(out, context);
       ws.sub();
       ws.println("}");
       ws.println("catch( Exception $throwable ) {" );
       ws.add();
       ws.println(catchVariableName + ".select($throwable);");
       ws.sub();
       ws.println("}");
       ws.println("finally {");
       ws.add();
       ws.println( runtimeString + "._$removeCatchObjectStack();");
       if ( finallyBlock != null ) {
           finallyBlock.gen(out, context);
       }
       ws.sub();
       ws.println("}");
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    private Expr expr;
    private StatementList statementList;
    private StatementList finallyBlock;

}
