package GreenCompiler;

import java.io.*;

public class CreateArray {
    static public void main( String []args ) {

    String directory = "";
    if ( args.length == 1 ) {
       directory = args[0] + "\\";
    }

    char []array = readFile( "array-1.g" );
    createArrays( array, directory, "$Array$1$char", "char", "_Char", "char" );
    createArrays( array, directory, "$Array$1$boolean", "boolean", "_Boolean", "boolean" );
    createArrays( array, directory, "$Array$1$byte", "byte", "_Byte", "byte" );
    createArrays( array, directory, "$Array$1$int", "integer", "_Integer", "int" );
    createArrays( array, directory, "$Array$1$long", "long", "_Long", "long" );
    createArrays( array, directory, "$Array$1$float", "real", "_Real", "float" );
    createArrays( array, directory, "$Array$1$double", "double", "_Double", "double" );

    /*
    array = readFile( "array-2.g" );
    createArrays( array, directory, "$Array$2$_char", "char", "_Char", "char" );
    createArrays( array, directory, "$Array$2$_boolean", "boolean", "_Boolean", "boolean" );
    createArrays( array, directory, "$Array$2$_byte", "byte", "_Byte", "byte" );
    createArrays( array, directory, "$Array$2$_integer", "integer", "_Integer", "int" );
    createArrays( array, directory, "$Array$2$_long", "long", "_Long", "long" );
    createArrays( array, directory, "$Array$2$_real", "real", "_Real", "float" );
    createArrays( array, directory, "$Array$2$_double", "double", "_Double", "double" );
    */

    // not used : @ is the array class name
    // ^ is the type of the array elements in Green
    // not used : # is the wrapper class name associated to ^ in Green
    // ? is the java name for the array element type

    array = readFile( "iter-1.g" );
    createArrays( array, directory, "BasicIter$char", "char", "_Char", "char" );
    createArrays( array, directory, "BasicIter$boolean", "boolean", "_Boolean", "boolean" );
    createArrays( array, directory, "BasicIter$byte", "byte", "_Byte", "byte" );
    createArrays( array, directory, "BasicIter$int", "integer", "_Integer", "int" );
    createArrays( array, directory, "BasicIter$long", "long", "_Long", "long" );
    createArrays( array, directory, "BasicIter$float", "real", "_Real", "float" );
    createArrays( array, directory, "BasicIter$double", "double", "_Double", "double" );


    /*
    array = readFile( "iter-2.g" );
    createArrays( array, directory, "$Iter$Array$2$_char", "char", "_Char", "char" );
    createArrays( array, directory, "$Iter$Array$2$_boolean", "boolean", "_Boolean", "boolean" );
    createArrays( array, directory, "$Iter$Array$2$_byte", "byte", "_Byte", "byte" );
    createArrays( array, directory, "$Iter$Array$2$_integer", "integer", "_Integer", "int" );
    createArrays( array, directory, "$Iter$Array$2$_long", "long", "_Long", "long" );
    createArrays( array, directory, "$Iter$Array$2$_real", "real", "_Real", "float" );
    createArrays( array, directory, "$Iter$Array$2$_double", "double", "_Double", "double" );
    */

    }

    static public void createArrays(
       char []array,
       String directory,  // with separator. Example:   c:\jose\
       String arrayClassName,
       String arrayElementType,
       String wrapperType,
       String javaElementType
       ) {

       PrintWriter out;
       try {
          out =
             new PrintWriter( new FileOutputStream( directory + arrayClassName + ".g" ) ) ;
       } catch ( Exception e ) {
           show("File " + directory + arrayClassName + ".g cannot be created" );
           return;
       }

       for (int j = 0; j < array.length; j++)
         switch ( array[j] ) {
            case ch_arrayClassName :
               out.print( arrayClassName );
               break;
            case ch_arrayElementType :
               out.print( arrayElementType );
               break;
            case ch_wrapperType :
               out.print( wrapperType );
               break;
            case ch_javaElementType :
               out.print( javaElementType );
               break;
            default :
               out.print(array[j]);
         }
       out.close();
       }

    static final char ch_arrayClassName = '@';
    static final char ch_arrayElementType = '^';
    static final char ch_wrapperType = '#';
    static final char ch_methodName = '%';
    static final char ch_javaElementType = '?';


// @ is the array class name
// ^ is the type of the array elements
// # is the wrapper class name associated to ^ in Green
// % is the method name of # that retrieves the value stored in the wrapper object


    static public char []readFile( String fileName ) {
        FileReader stream;
        int numChRead;

        File file = new File(fileName);
        if ( ! file.exists() || ! file.canRead() ) {
           show("Either the file " + fileName + " does not exist or it cannot be read");
           return null;
         }
         try {
             stream = new FileReader(file);
         } catch ( FileNotFoundException e ) {
            show("Something wrong: file " + fileName + " does not exist anymore");
            return null;
         }
                // one more character for '\0' at the end that will be added by the
                // compiler
         char []input = new char[ (int ) file.length() + 1 ];

         try {
            numChRead = stream.read( input, 0, (int ) file.length() );
         } catch ( IOException e ) {
            show("Error reading file " + fileName);
            return null;
         }

         if ( numChRead != file.length() ) {
            show("Read error in file " + fileName);
            return null;
         }
         try {
            stream.close();
         } catch ( IOException e ) {
            show("Error in handling the file " + fileName );
            return null;
         }
         return input;
    }

    static private void show( String msg ) {
        System.out.println(msg);
        System.exit(0);
    }

}



