package GreenCompiler;

import AST.*;
import Interface.*;
import java.util.*;

public class NameServer {

    
   public static String getSearchWrapperName( ExprList anExprList ) {
    
      String searchWrapper = "";
      Enumeration e = anExprList.elements();
      while ( e.hasMoreElements() ) {
         Type t = ((Expr ) e.nextElement()).getType();
         if ( t instanceof BasicType ) 
            searchWrapper += ((BasicType ) t).getWrapperName() + "$";
         else 
            searchWrapper += t.getName() + "$";
      }
      return searchWrapper;
   }

   
   public static String getSearchWrapperName( ParameterList paramList ) {
    
      String searchWrapper = "";
      Enumeration e = paramList.elements();
      while ( e.hasMoreElements() ) {
         Type t = ((Parameter ) e.nextElement()).getType();
         if ( t instanceof BasicType ) 
            searchWrapper += ((BasicType ) t).getWrapperName() + "$";
         else 
            searchWrapper += t.getName() + "$";
      }
      return searchWrapper;
   }
   
   
   public static String getClassMethodInfoNameVar( String methodName, ParameterList paramList ) {
      // return the name of the variable that holds an object of
      // ClassMethodInfo with info of the method methodName.
      return "$cmi$" + methodName + NameServer.buildParameterListName(paramList);
   }
   
   public static String getObjectMethodInfoNameVar( String methodName, ParameterList paramList ) {
      // return the name of the variable that holds an object of
      // ObjectMethodInfo with info of the method methodName.
      return "$omi$" + methodName + NameServer.buildParameterListName(paramList);
   }
   
   public static String getMethodJavaName( String methodName, ParameterList paramList ) {
      // return the name of the method in Java

      char ch = methodName.charAt(0);
      if ( Character.isLetter(ch) || ch == '$' || ch == '_' )
         return "_" + methodName + NameServer.buildParameterListName(paramList);
      else 
         // should be an operator
         return NameServer.getOperatorName(methodName);
    
   }
   
   public static String getMethodInterfaceName( String methodName, Type returnType,
          ParameterList paramList ) {
      // return the name of the interface corresponding to the
      // method
            
      char ch = methodName.charAt(0);
      if ( ! Character.isLetter(ch) && ch != '$' && ch != '_' )
         methodName = NameServer.getOperatorName(methodName);
            
      String name = "$mi$" + methodName + "$" + paramList.getSize() + 
          NameServer.buildParameterListName(paramList); 
      if ( returnType != null ) {
//          if ( returnType instanceof BasicType || GreenProject.subtypeEqualSubclass )
             name = name + "$" + returnType.getJavaName();
//          else 
//             name = name + "$_Any";
      }
      return name;
   }
   
   public static String getParameterNameVar( String paramName ) {
      // return the name of a variable corresponding the the 
      // parameter paramName
      return "$p$" + paramName;
   }
   
   public static String getOperatorName( String stringOperator ) {
      // return the name of the operator stringOperator
       return (String ) operatorNameTable.get(stringOperator);
   }
 
   public static String getAccessToClassObject( String name ) {
      /*
         return a string with the code that accesses the 
         class object of class name. If name is "A", this method
         should return _A.$classObject
      */
      String result = 
        (name.compareTo("String") == 0 ? "_String" : NameServer.buildRealClassJavaName(name)) + ".$classObject";
      /* if ( name.compareTo("String") == 0 )
         System.out.println(); */
      return result;
   }
   
   public static String getAccessToClassInfoClass( String javaName ) {
      // return a string with code that accesses the 
      // object with info on the class. javaName should be the
      // class name in Java.
      return javaName + ".$i";
   }
   
   public static String getAccessToClassInfoClassObject( String javaName ) {
      // return a string with code that accesses the 
      // object with info on the class object. javaName should be the
      // class name in Java.
      return javaName + ".$i";
   }
 
   public static String getShellClassName( String shellName, ClassType attached ) {
      // return the name of the shell class. This class is that with
      // the methods
      return "$shell$" + NameServer.buildJavaName(shellName) + "$" + 
         attached.getJavaName();
   }
   
   public static String getShell_ivClassName( String shellJavaName ) {
      // return the name of the shell class. This class is that with
      // the instance variables.
      return "$shellivc$" + shellJavaName;
   }
   
   public static String getJavaFileName(String name, String javaName) {
      // name is the name of the class in Green. javaName is the
      // name of the class in java. This method returns the name
      // of the file in which the class is. Without the ".java"
      if ( basicTypes.get(name) == null ) 
         return javaName;
      else
         return "__" + name;
   }
   
   public static Hashtable operatorNameTable;   
   static {
      operatorNameTable = new Hashtable();
      operatorNameTable.put("and", "and");
      operatorNameTable.put("not", "not");
      operatorNameTable.put("or", "or");
      operatorNameTable.put("xor" , "xor");
      operatorNameTable.put("&", "binand");
      operatorNameTable.put("~", "binnot");
      operatorNameTable.put("|", "binor");
      operatorNameTable.put("^" , "binxor");
      operatorNameTable.put(">>", "rightshift");
      operatorNameTable.put("<<", "leftshift");
      operatorNameTable.put("==", "eq");
      operatorNameTable.put("=" , "assign");
      operatorNameTable.put("<>", "neq");
      operatorNameTable.put(">=", "ge");
      operatorNameTable.put(">" , "gt");
      operatorNameTable.put("<=", "le");
      operatorNameTable.put("<" , "lt");
      operatorNameTable.put("++", "plusplus");
      operatorNameTable.put("--", "minusminus");
      operatorNameTable.put("+" , "plus");
      operatorNameTable.put("-" , "minus");
      operatorNameTable.put("*" , "mult");
      operatorNameTable.put("/" , "div");
      operatorNameTable.put("%" , "remainder");
   }
   
   
    public static String buildMethodSearchName(String methodName, ExprList exprList ) {
        // returns a string used to search for a method - takes polymorphism in account.
        
        String methodSearchName = methodName + "$" + exprList.getSize();
        Enumeration e = exprList.elements();
        while ( e.hasMoreElements() ) {
            Expr expr = (Expr ) e.nextElement();
            /*m1
            methodSearchName = methodSearchName + "$" + expr.getType().getJavaName();
            */
            Type t = expr.getType();
            if ( t instanceof BasicType ) 
               methodSearchName += "$" + ((BasicType ) t).getJavaWrapperName();
            else 
               methodSearchName += "$" + t.getJavaName();
            
        }
        return methodSearchName;
    }
    

    public static String buildMethodSearchName(String methodName, ParameterList paramList) {
        // returns a string used to search for a method - takes polymorphism in account.
        
        String methodSearchName = methodName + "$" + 
            paramList.getSize() + buildParameterListName(paramList);
        return methodSearchName;
    }
    
    public static String buildParameterListName( ParameterList paramList ) {
        // builds a String with the name of the parameters. Used to,
        // for example, build the name of a method.
        
        String paramListName = "";
        Enumeration e = paramList.elements();
        while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            /*m1
            paramListName = paramListName + "$" + p.getType().getJavaName();
            */
            Type t = p.getType();
            if ( t instanceof BasicType ) 
               paramListName += "$" + ((BasicType ) t).getJavaWrapperName();
            else 
               paramListName += "$" + t.getJavaName();
            
        }
        return paramListName;
    }


    public static String buildAnySearchName(String methodName, ExprList exprList ) {
        // Put Any in place of the names of classes. So  "proc m(real, Person)"
        // results in  m$2$float$_Any
        // returns a string used in the search for a method name.
        
        String methodSearchName = methodName + "$" + exprList.getSize();
        Enumeration e = exprList.elements();
        while ( e.hasMoreElements() ) {
            Expr expr = (Expr ) e.nextElement();
            /*m1
            if ( expr.getType() instanceof BasicType ||
                 GreenProject.subtypeEqualSubclass ||
                 expr.getType().getIsJavaClass() ) {
            */
               /*m1
               methodSearchName = methodSearchName + "$" + expr.getType().getJavaName();
               */
            /*m1
               Type t = expr.getType();
               if ( t instanceof BasicType ) 
                  methodSearchName += "$" + ((BasicType ) t).getJavaWrapperName();
               else 
                  methodSearchName += "$" + t.getJavaName();
            }
            else
            */
               methodSearchName = methodSearchName + "$_Any";
        }
        return methodSearchName;
    }
    

    public static String buildAnySearchName(String methodName, ParameterList paramList) {
        // Put Any in place of the names of classes. So  "proc m(real, Person)"
        // results in  m$2$float$_Any
        // returns a string used in the search for a method name.
        
        String methodSearchName = methodName + "$" + paramList.getSize() +
           buildAnyParameterListName(paramList);
        return methodSearchName;
    }

    public static String buildAnyParameterListName( ParameterList paramList ) {
        /* 
          returns a string with the parameters types replaced by
          Any when they are non-basic types. used by 
          buildAnySearchName
        */
        String s = "";
        Enumeration e = paramList.elements();
        while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            /*m1
            if ( p.getType() instanceof BasicType || 
                 GreenProject.subtypeEqualSubclass ||
                 p.getType().getIsJavaClass() ) {
            */
               /*m1
               s = s + "$" + p.getType().getJavaName();
               */
            /*m1
               Type t = p.getType();
               if ( t instanceof BasicType ) 
                  s += "$" + ((BasicType ) t).getJavaWrapperName();
               else 
                  s += "$" + t.getJavaName();
            }
            else
            */
               s = s + "$_Any";
        }
        return s;
    }

    
    public static String buildClassName( String className, ExprList exprList ) {
        /*
          returns a string with the class name. The parameter
          className is the name of the class. exprList is a list
          of expressions in case of an instantiation of a 
          parameterized class. For a class List(char) the return
          is $pc$List$1$char. If exprList is null, the result
          is className.
        */
        if ( exprList != null && exprList.getSize() > 0 ) {
           className = "$pc$" + className + "$" + exprList.getSize();
           Enumeration e = exprList.elements();
           while ( e.hasMoreElements() ) {
               Expr expr = (Expr ) e.nextElement();
               if ( ! (expr instanceof ClassObjectExpr) )
                  return null;
               Type t = ((ClassObjectExpr ) expr).getClassObjectType();
               if ( ! (t.getTypeInterface() instanceof ClassObjectTypeInterface) ) 
                  return null;
               ClassObjectTypeInterface coti = (ClassObjectTypeInterface ) t.getTypeInterface();
               className = className + "$" + coti.getClassTypeInterface().getJavaName();
           }
        }
        return className;
    }
    

    public static String buildClassName( String className, TypeList typeList ) {
        /*
          returns a string with the class name. The parameter
          className is the name of the class. typeList is a list
          of types in case of a parameterized class. For a class 
          List with typeList containing char, the return is 
          $pc$List$1$char. If typeList is null, the result
          is className.
        */
        if ( typeList != null && typeList.getSize() > 0 ) {
           className = "$pc$" + className + "$" + typeList.getSize();
           Enumeration e = typeList.elements();
           while ( e.hasMoreElements() ) 
               className = className + "$" + ((Type ) e.nextElement()).getJavaName();
        }
            
        return className;
    }
    
    public static String buildClassObjectName( String className ) {
        // returns the name of the class object of class className,
        // in which className is the name of the class in Green.
        if ( basicTypes.get(className) != null ) 
           return "$co$_" + className;
        else
           return "$co$" + className;
    }
    
    public static String buildClassObjectName( TypeInterface cti ) {
        // returns the name of the class object of class cti,
        // in which cit is the class interface of the class in Green.
        return buildClassObjectName(cti.getName());
    }
    
    public static String buildArrayName( Type type, int dim ) {
        // returns the name of an array whose type is "type" and
        // has dim dimensions. 
        return "$Array$" + dim + "$" + type.getJavaName();
    }
    
    public static String buildArrayName( String javaName, int dim ) {
        // returns the name of an array whose type is javaName and
        // has dim dimensions. 
        return "$Array$" + dim + "$" + javaName;
    }
    
    public static String buildJavaName( String name ) {
        /* 
          returns the name in Java of class "name". Then, if 
          name is "integer", the return will be "int". If name
          is "$pc$List$1$char" the return will be "$pc$List$1$char".
          If name is "A", the result will be "_A".
        */
        
       char ch = name.charAt(0);
       if ( ch == '$' || ch == '_' )
          return name;
       else {
          String s;
          if ( (s = (String ) basicClassTable.get(name)) != null ) 
             return s;
          else 
             return "_" + name;
       }
    }
    
    public static String buildRealClassJavaName( String name ) {
        /*
           returns the name of a class in Java. It calls 
           buildJavaName for non-basic types.
           For the basic types, it returns something like 
           __integer or __char.
        */
        if ( basicTypes.get(name) == null )
           return buildJavaName(name);
        else
           return "__" + name;
    }
        
    
    public static String buildBasicTypeClassName( String name ) {
       // name is the name of a basic type. If String, the 
       // return value will be "_String". Otherwise, this method
       // will return something like "__integer".
       if ( name.compareTo("String") == 0 )
          return "_"  + name;
       else 
          return "__"  + name;
    }


    static public Hashtable basicClassTable, basicTypes;
    static {
        basicClassTable = new Hashtable();
        basicClassTable.put("char", "char");
        basicClassTable.put("byte", "byte");
        basicClassTable.put("boolean", "boolean");
        basicClassTable.put("integer", "int");
        basicClassTable.put("long", "long");
        basicClassTable.put("real", "float");
        basicClassTable.put("double", "double");
        basicClassTable.put("String", "String");

        basicTypes = new Hashtable();
        basicTypes.put("char", "");
        basicTypes.put("byte", "");
        basicTypes.put("boolean", "");
        basicTypes.put("integer", "");
        basicTypes.put("long", "");
        basicTypes.put("real", "");
        basicTypes.put("double", "");
    
    }
    
   
}