package GreenCompiler;

import java.util.*;
import java.io.*;
import AST.*;
import Interface.*;


public class ParserParamClass {

    public ParserParamClass(
               ProjectError error,
               String projectDir,
               String classDir,
               String separator
               ) {

       this.error= error;
       this.projectDir = projectDir;
       this.classDir = classDir;
       this.separator = separator;
    }

    private Hashtable table;
    private char []in;
    private ProjectError error;
    private PrintWriter out;

    private int lineNumber, beginCurrentLine;
    private String ident;
    private int k;
    private int offset;
    private int token;
    private boolean inJavaCode;
    private String projectDir;
    private String classDir;
    private String separator;
    private String newClassName;


    static private final int
       EOF = -1,
       IDENT = 0,
       JAVABEGIN = 1,
       JAVAEND = 2,
       SPACE = 3,
       CLASS = 4,
       OBJECT = 5,
       CLASSNAME = 6;


    public static void main( String []args ) {
        (new ParserParamClass( 
            new ProjectError(new PrintWriter(System.out)),
            "c:\\jose\\green\\compiler\\rec", "test", "\\")).start();
    }
    
    
    

    public ClassFile create( ClassType classType, ParameterizedClassInterface aClass )
       /* create a real class from classType, which is something like List(char) or Stack(Person)
          aClass is the object representing the parameterized class 
          Return an object ClassFile with the directory and name of the created class.
          */
    {

        ClassFile cf = aClass.getClassFile();
        String pcFileName;
        pcFileName = cf.getDirectory() + separator + cf.getClassName() + ".g";

        //@ ClassFile outClassFile = new ClassFile(projectDir + separator + classDir, classType.getJavaName() );
        ClassFile outClassFile = new ClassFile(classDir, classType.getJavaName() );
        
        String outName = outClassFile.getDirectory() + separator + outClassFile.getClassName() + ".g";


        File fileIn, fileOut;
        fileIn = new File(pcFileName);
        fileOut = new File(outName);
        if ( fileIn.lastModified() < fileOut.lastModified() )
              // parameterized class has not been modified since the instantiated 
              // parameterized class was created
           return outClassFile;
           
        newClassName = classType.getJavaName();

        table = new Hashtable();
        table.put( "java$begin", new Integer(JAVABEGIN) );
        table.put( "java$end", new Integer(JAVAEND) );
        table.put( "class", new Integer(CLASS) );
        table.put( "object", new Integer(OBJECT) );
        table.put( classType.getName(), new Integer(CLASSNAME) );
        

        ClassParameter formalParameter;
        Type realParameter;

        Enumeration formalList = aClass.getParamList().elements();
        Enumeration realList   = classType.getTypeList().elements();

        if ( aClass.getParamList().size() != classType.getTypeList().getSize() ) {
           error.show("Parameterized class " + aClass.getName() + " used with wrong number of parameters");
           return null;
        }
        while ( formalList.hasMoreElements() ) {
            formalParameter = (ClassParameter ) formalList.nextElement();
            realParameter = (Type ) realList.nextElement();
            table.put( formalParameter.getName(),
                new RealFormalClassParameter(realParameter.getName(), realParameter.getJavaName() ) );
        }

        
        MyFile myfile = new MyFile(pcFileName);
        in = myfile.readFile();
        if ( myfile.getError() != MyFile.ok_e ) {
            error.show("Error reading file " + pcFileName);
            return null;
        }

           // put white space in place of the class parameters
        int last = aClass.getEndClassParam();
        for( int j = aClass.getStartClassParam(); j < last; j++ )
           in[j] = ' ';


        try {
            out = new PrintWriter( new FileOutputStream(outName) );
        } catch ( Exception e ) {
            error.show("Error creating file " + outName);
            return null;
        }
        
       
        initLexer();
        parse();
        out.close();
        
        return outClassFile;
    }





    private void parse() {

        while ( token != EOF ) {
            switch ( token ) {
               case SPACE :
                  out.write(in, offset, k - offset);
                  break;
               case IDENT :
               case CLASSNAME :
                  out.print(ident);
                  break;
               case JAVABEGIN :
                  out.print(ident);
                  inJavaCode = true;
                  break;
               case JAVAEND :
                  out.print(ident);
                  inJavaCode = false;
                  break;
               case CLASS :
               case OBJECT :
                  out.print(ident);
                  next();
                  if ( token != SPACE ) 
                     error.show("Internal error at ParserParamClass");
                  else {
                     out.write(in, offset, k - offset);
                     next();
                        // if token != IDENT, there is a internal error, for an error
                        // should be displayed when contructing the class interface
                     if ( token == CLASSNAME )
                        out.print(newClassName);
                     else 
                        error.show("Internal error at ParserParamClass");
                  }
            }
            next();
        }
        if ( out.checkError() ) {
            error.show("Error while writing to file");
        }
    }




    public void initLexer() {
       k = 0;
       beginCurrentLine = 0;
       lineNumber = 1;
       token = EOF;
       inJavaCode = false;
       next();
    }


    public void next() {
        offset = k;
        while ( Character.isWhitespace(in[k]) ) {
           if ( in[k] == '\n' ) {
              lineNumber++;
              beginCurrentLine = k + 1;
           }
           k++;
        }
        if ( k > offset ) {
           token = SPACE;
           return;
        }
        if ( in[k] == '\0' ) {
           token = EOF;
           return;
        }


        if ( in[k] == '/' && in[k + 1] == '/' ) {
            // comment till the end of the line
            k += 2;
            while ( in[k] != '\n' && in[k] != '\0' )
               k++;
            if ( in[k] == '\n' ) {
               k++;
               lineNumber++;
               beginCurrentLine = k;
            }
            token = SPACE;
            //next();
        }
        else
            if ( in[k] == '/' && in[k + 1] == '*' ) {
                // start comment
                int commentLevel = 1;
                int lineStartComment = lineNumber;
                k += 2;
                while ( commentLevel > 0 )
                    if ( in[k] == '*' && in[k + 1] == '/' ) {
                        commentLevel--;
                        k += 2;
                    }
                    else if ( in[k] == '/' && in[k + 1] == '*' ) {
                        commentLevel++;
                        k += 2;
                    } else if ( in[k] == '\0' ) {
                        error.show("Comment started at line " + lineStartComment + " does not terminate");
                        token = MySym.EOF;
                        return;
                    } else {
                        if ( in[k] == '\n' ) {
                           lineNumber++;
                           beginCurrentLine = k + 1;
                        }
                        k++;
                    }
                //next();
                token = SPACE;
            }
            else 
                // not a comment
                if ( Character.isLetter(in[k]) || in[k] == '_' || in[k] == '$' ) {
                    StringBuffer s = new StringBuffer();
                    while ( Character.isLetterOrDigit(in[k]) || in[k] == '_' || in[k] == '$' ) {
                       s.append( in[k] );
                       k++;
                    }
                    ident = s.toString();
                    Object v = table.get(ident);
                    if ( v == null ) {
                       // not java$begin, java$end, class parameter
                       token = SPACE;
                    }
                    else {
                       if ( v instanceof Integer ) {
                          token = ((Integer) v).intValue();
                       }
                       else {
                          // class parameter
                          RealFormalClassParameter cp = (RealFormalClassParameter ) v;
                          if ( inJavaCode )
                             ident = cp.getJavaName();
                          else
                             ident = cp.getGreenName();
                          token = IDENT;
                       }
                    }
                    return;
                }
                else 
                   switch ( in[k] ) {
                      case '\"' : {
                         k++;
                         while ( in[k] != '\0' && in[k] != '\n' )
                            if ( in[k] == '\"' )
                               break;
                            else
                               if ( in[k] == '\\' ) {
                                  if ( in[k+1] == '\\' ) {
                                      k += 2;
                                  }
                                  else
                                     if ( in[k+1] != '\n' && in[k+1] != '\0' ) {
                                        k += 2;
                                     }
                               }
                               else {
                                  k++;
                               }

                         if ( in[k] == '\0' || in[k] == '\n' ) {
                            error.show("Nonterminated string");
                         }
                         else {
                            k++;
                         }
                         token = SPACE;
                         }
                         break;
                      case '\'' : {
                         k++;
                         if( in[k] == '\\' )
                            while ( in[k] != '\'' && in[k] != '\n' && in[k] != '\0' ) {
                                k++;
                            }
                         else {
                            k++;
                         }
                         if ( in[k] != '\'' ) {
                            if ( in[k] == '\n' || in[k] == '\0' )
                               error.show("Non-terminated literal character");
                            else
                               error.show("\' expected");
                         }
                         else
                            k++;
                         }
                         token = SPACE;
                         break;
                      default :

                         k++;
                         token = SPACE;
                   }
                
    }
    
    
    public void start() {
        
        table = new Hashtable();
        table.put( "java$begin", new Integer(JAVABEGIN) );
        table.put( "java$end", new Integer(JAVAEND) );
        table.put( "class", new Integer(CLASS) );
        table.put( "object", new Integer(OBJECT) );
        table.put( "ArrayVector", new Integer(CLASSNAME) );
        table.put( "T", 
          new RealFormalClassParameter( "Person", "_Person" ) );

        String pcFileName;
        pcFileName = "c:\\jose\\green\\compiler\\test1\\ArrayVector.g";
        
        MyFile myfile = new MyFile(pcFileName);
        in = myfile.readFile();
        if ( myfile.getError() != MyFile.ok_e ) {
            error.show("Error reading file " + pcFileName);
            return;
        }

        out = null;
        String outName = "c:\\jose\\green\\compiler\\test1\\pc$ArrayVector$Person.g";

        try {
            out = new PrintWriter( new FileOutputStream(outName) );
        } catch ( Exception e ) {
            error.show("Error creating file " + outName);
            return;
        }

       initLexer();
       parse();

    }
    
    
}
