package GreenCompiler;

import java.util.*;
import AST.*;
import Interface.*;


public class SymbolTable {
    
    public SymbolTable() {
        localTable = new Hashtable();
        classTable = new Hashtable();
        tryTypeList = new Vector();
    }
    
    public Identifier putInLocal( Identifier id ) {
        return (Identifier ) localTable.put(id.getSearchName(), id);
    }
    
    public Identifier putInClass( Identifier id ) {
        // add identifier to a list containing the class methods and instance variables
        return (Identifier ) classTable.put(id.getSearchName(), id);
    }
    
    
    public Identifier get( String name ) {
        Identifier obj;
        
        if ( (obj = (Identifier ) localTable.get(name)) != null )
          return obj;
        else 
          return (Identifier ) classTable.get(name);
    }
    
    public Identifier getInLocal( String name ) {
        return (Identifier ) localTable.get(name);
    }
    
    public Identifier getInClass( String name ) {
        return (Identifier ) classTable.get(name);
    }
        
    public Identifier removeLocal( String name ) {
       return (Identifier ) localTable.remove(name);
    }
    
    public void deleteLocalIdent() {
       
       tryTypeList.removeAllElements();
       if ( localTable != null && localTable.size() > 0 )
          localTable = new Hashtable();
    }
    
    public void deleteTryList() {
        tryTypeList.removeAllElements();
    }
    
    
    public void deleteAll() {
        deleteLocalIdent();
        classTable = new Hashtable();
        if ( tryTypeList != null && tryTypeList.size() > 0 )
           tryTypeList = new Vector();
    }
    
    
    
    public void putTryType( Type catchType ) {
        tryTypeList.addElement(catchType);
    }
    
    public void removeTryType() {
        tryTypeList.removeElementAt( tryTypeList.size() - 1 );
    }
    
    
    public MethodInterface getMethodOfCatch( String methodSearchName ) {
        
       Type t;
       int i = tryTypeList.size() - 1;
       while ( i >= 0 ) {
          t = (Type ) tryTypeList.elementAt(i);
          MethodInterface mi = t.getTypeInterface().publicSearch(methodSearchName);
          if ( mi != null ) 
             return mi;
          i--;
       }
       return null;
    }
    
    public Hashtable multipleGetMethodOfCatch( String anyMethodSearchName ) {
        
       Hashtable result = new Hashtable();
       Type t;
       int i = tryTypeList.size() - 1;
       while ( i >= 0 ) {
          t = (Type ) tryTypeList.elementAt(i);
          t.getTypeInterface().multiplePublicSearch(anyMethodSearchName, result);
          i--;
       }
       return result;
    }
    
    
    
    private Hashtable localTable, classTable;
    private Vector tryTypeList;
    
}