object ArrayVector( T : Any )
   public:
      const defaultSizeArrayVector = 20;
      const defaultAddArrayVector = 10;
end    
   
  
  // since object ArrayVector has a parameter, class ArrayVector must have too.
  
class ArrayVector subclassOf  Container(T)

    proc init( maxSize : integer )
         /* maximum size of the ArrayVector. Initially the number of elements is
            zero. That is,  getSize() returns 0. */
       begin
       self.maxSize = maxSize;
       size = 0;
       index = -2;
       v = array(T)[].new(maxSize);
       end
        
    proc init()
         // uses defaultSizeArrayVector to initiate the object
      begin
      init(ArrayVector(T).defaultSizeArrayVector);
      end
      
  public:
    
    proc add( elem : T )
            ( exception : CatchOutOfMemoryException )
         // inserts elem after the last array element
       begin
       if size >= maxSize
       then
          if not expandArray()
          then
             exception.throw( OutOfMemoryException.new() );
          endif
       endif
       v[size] = elem;
       ++size;
       end
             

    proc get() ( exception : CatchNotFoundException ) : T
          /*  gets one element from the container which is not removed
              from it. */
       begin
       if size > 0
       then
          return v[size-1];
       else
          exception.throw( NotFoundException.new() );
          // return nil;
       endif
       end
                             
                             
    proc remove() ( exception : CatchNotFoundException ) : T
         // removes one element from the container and returns it.
       begin
       if size > 0
       then
          var elem : T = v[size-1];
          --size;
          return elem;
       else
          exception.throw( NotFoundException.new() );
          // return nil;
       endif
       end
       

    proc removeAll()
          // empties the container
       begin
       size = 0;
       end
        
        
    proc getSize() : integer
         // returns the real number of ArrayVector elements
       begin
       return size;
       end

    proc empty() : boolean
          // returns true if there is no element in the container
       begin
       return size == 0;
       end
      
      
      
    proc full() : boolean
         /* returns true if there is no more space in the container. This
            method only returns false in static containers whose number of
            elements is defined at the creation of the container */
       begin
       return true;
       end

      
    proc getIter() : Iter(T)
          // returns an iterator for the container
          var vi : ArrayVectorIter(T);
       begin
       vi = ArrayVectorIter(T).new(v, size);
       return vi;
       end
      
      
    proc forEach( f : Function(T) )
          // calls "f.exec(x)" for each element x of the container
          var i : integer;
       begin
       for i = 0 to size - 1 do
          f.exec(v[i]);
       end
       

      
    proc collect( f : Filter(T) ) : Container(T)
          /* collects all container elements x such that "f.test(x)"
             evaluates to true. These elements are inserted in a new
             container that is returned by this method. */
          var newArrayVector : ArrayVector(T);
              i : integer;
       begin
       newArrayVector#init(size);
       for i = 0 to size - 1  do
          if f.test(v[i]) 
          then
             try(HCatchAll)
                newArrayVector.add(v[i]);
             end
          endif
       return newArrayVector;
       end


    proc reset()
       begin
       if index <> -2
       then
          Out.writeln("Iterator error");
          Runtime.exit(1);
       else
          index = -1;
       endif
       end
  
    proc next()
             ( exception : CatchNotFoundException ) : T
       begin
       ++index;
       if index == size
       then
          index = -2;
          exception.throw( NotFoundException.new() );
       else
          return v[index];
       endif
       end
      
    proc endIter()
      begin
      index = -2;
      end

      
    proc toArray() : array(T)[]
          var newV : array(T)[];
       begin
       newV = array(T)[].new(size);
       for i : integer = 0 to size - 1 do
          newV[i] = v[i];
       return newV;
       end
    

  subclass:
  
    proc s_getMaxSize() : integer
         // returns the maximum number of ArrayVector elements
       begin
       return maxSize;
       end
      
   
    proc s_add( elem : T; i : integer )
            ( exception : CatchOutOfMemoryException )
         // inserts elem in the position i. See observations above.
       begin
       if i >= size
       then
          for j : integer = size to i - 1 do
             add(nil);
       endif
       // add(elem);
       v[size] = elem;
       end
            
    proc s_get( i : integer )
            ( exception : CatchNotFoundException ) : T
          /* gets the i-th element from the ArrayVector which is not removed
             from it. */
       begin
       if i >= size
       then
          exception.throw( NotFoundException.new() );
          // return nil;
       else
          return v[i];
       endif
       end
       
    proc s_remove( i : integer )
                 ( exception : CatchNotFoundException )  : T
          /* removes the i-th element from the ArrayVector and returns it.  All
             other elements are left-shifted. Exception NotFoundException is
             thrown if i < 0 or i >= getNum() */
       begin
       if i < size 
       then
          var elem : T = v[i];
          --size;
          for j : integer = i to size - 1 do
             v[j] = v[j+1];
          return elem;
       else
          exception.throw( NotFoundException.new() );
          // return nil;
       endif
       end

    proc  s_remove( f : Filter(T) )
          /* removes all array elements x from the container whenever
             "f.test(x)" evaluates to true. */
          var w : array(T)[];
              i, j : integer;
       begin
       w#init(maxSize);
       j = 0;
       for i = 0 to size - 1  do
          if not f.test(v[i]) 
          then
             w[j] = v[i];
             ++j;
          endif
       size = j;
       v = w;
       end

    proc s_replaceBy( cmd : Command(T) )
          // replaces each element x in the container by "cmd.doIt(x)"
          var i : integer;
       begin
       for i = 0 to size - 1 do
          v[i] = cmd.doIt(v[i]);
       end
      


      proc getIndex() : integer
         begin
         return index;
         end
         
      proc setIndex( index : integer )
         begin
         self.index = index;
         end
   
   private:
   
      var size, maxSize, index : integer;   
          v : array(T)[];
          
      proc expandArray() : boolean
            var newV : array(T)[];
                i    : integer;
         begin
         CatchAll.initialize();
         newV = nil;
         try(CatchAll)
            newV = array(T)[].new( maxSize + ArrayVector(T).defaultAddArrayVector );
         end
         if CatchAll.getException() <> nil
         then
            return false;
         else
            maxSize = maxSize + ArrayVector(T).defaultAddArrayVector;
            for i = 0 to size - 1 do
               newV[i] = v[i];
            v = newV;
            return true;
         endif
         end
      
end      
