abstract class Container( T : Any )

    proc init()
      begin
      end
      
  public:
  
    abstract proc add( elem : T )
                     ( exception : CatchOutOfMemoryException )
       // inserts elem in the container
       
    abstract proc get() 
                     ( exception : CatchNotFoundException ) : T
       /* gets one element from the container which is not removed
          from it. */
                             
    abstract proc remove()
                     ( exception : CatchNotFoundException ) : T 
      // removes one element from the container and returns it.

    abstract proc removeAll()
      // empties the container
                                                      
    abstract proc getSize() : integer
      // returns the number of elements in the container
      
    abstract proc empty() : boolean
      // returns true if there is no element in the container
      
    abstract proc full() : boolean
      /* returns true if there is no more space in the container. This
         method only returns false in static containers whose number of
         elements is defined at the creation of the container */
      
    abstract proc getIter() : Iter(T)
      // returns an iterator for the container
      
    abstract proc forEach( f : Function(T) )
      // calls "f.exec(x)" for each element x of the container

    abstract proc collect( f : Filter(T) ) : Container(T)
      /* collects all container elements x such that "f.test(x)"
         evaluates to true. These elements are inserted in a new
         container that is returned by this method. */

    abstract proc reset()
      /* reset, next, and endIter compose a lightWeight iterator. It
         should be used as in the example:
            list.reset();
            while (elem = list.next()) <> nil do
               Out.writeln(elem);
            list.endIter();
         This iterator can only be used when T is a reference class
         (not a basic class like char, byte, etc) and nil cannot be
         inserted in the container. Method next returns nil when there is no
         more elements.
            Method endIter should be called after the use of the
         iterator. If it is not called, the next call to reset will cause a
         run-time error:
             list.reset();
             x = list.next();
               // run-time error in the following line
             list.reset();
         This prevents one of calling reset in the middle of an use of
         the iterator:
             list.reset();
             while (x = list.next()) <> nil do
               begin
               select( list, x );
               ...
               end
         If select calls ``reset'' on list, there will be an
         error. Only one scanning of the container can be made at a time with
         this kind of iterator. 
         
           This type of iterator does not need a separate object of
         type DS.Iter(T) thus saving memory and time of the garbage
         collector.
         
         This method should be implemented as
          begin
          if index <> -2
          then
             Out.writeln("Iterator error");
             Runtime.exit(1);
          else
             index = -1;
          endif
          end
      */
  
    abstract proc next() : T
      
    abstract proc endIter()

end
