abstract class DDoubleQueue( T : Any ) subclassOf Container(T)
  /* A double queue of elements. Insertion and removal of elements can
     occur in both sides of the queue.
     This class is better when the number of elements 
     varies greatly during the container lifetime. */
  
  public:
    // the same methods as DoubleQueue
       
    abstract proc add( elem : T )
                     ( exception : CatchOutOfMemoryException )
      // inserts elem in the container
      
    abstract proc get()
                     ( exception : CatchNotFoundException ) : T
      /* gets one element from the container which is not removed
        from it. */
                             
    abstract proc remove()
                     ( exception : CatchNotFoundException ) : T
      // removes one element from the container and returns it.

    abstract proc removeAll()
      // empties the container
                                                      
    abstract proc getSize() : integer
      // returns the number of elements in the container
      
    abstract proc empty() : boolean
      // returns true if there is no element in the container
      
    abstract proc full() : boolean
      /* returns true if there is no more space in the container. This
         method only returns false in static containers whose number of
         elements is defined at the creation of the container */
      
    abstract proc getIter() : Iter(T)
      // returns an iterator for the container
      
    abstract proc forEach( f : Function(T) )
      // calls "f.exec(x)" for each element x of the container

    abstract proc replaceBy( cmd : Command(T) )
      // replaces each element x in the container by "cmd.doIt(x)"
      
    abstract proc collect( f : Filter(T) ) : Container(T)
      /* collects all container elements x such that "f.test(x)"
         evaluates to true. These elements are inserted in a new
         container that is returned by this method. */

        
    proc reset()
      begin
      if index <> -2
      then
        Out.writeln("Iterator error");
        Runtime.exit(1);
      else
        index = -1;
      endif
      end
  
    abstract proc next() : T
      
    proc endIter()
      begin
      index = -2;
      end

end

