class DList( T : Any ) subclassOf Container(T)
  /* A list of elements. There is no order to insert or remove
     elements. 
     This class is better when the number of elements 
     varies greatly during the container lifetime. */
     
   proc init()
      begin
      first = nil;
      size = 0;
      end
  
   public:
  
      proc add( elem : T )
                     ( exception : CatchOutOfMemoryException )
         begin  
         var no : ListNo = ListNo.new(elem);
         no.setSucc(first);
         first = no;
         ++size;
         end
         
         
      proc get()
              ( exception : CatchNotFoundException ) : T
         begin
         if first == nil
         then
            exception.throw( NotFoundException.new() );
         else
            return first.getValue();
         endif
         end
                              
      proc remove()
                     ( exception : CatchNotFoundException )
         begin
         if first == nil
         then
            exception.throw( NotFoundException.new() );
         else
            var x : T = first.getValue();
            first = first.getSucc();
            --size;
            return x;
         endif
         end
                              

      proc removeAll()
         // empties the container
         begin
         first = nil;
         size = 0;
         end
                                                      
      proc getSize() : integer
         // returns the number of elements in the container
         begin
         return size;
         end
      
      proc empty() : boolean
         // returns true if there is no element in the container
         begin
         return first == nil;
         end
      
      proc full() : boolean
      /* returns true if there is no more space in the container. This
         method only returns true in static containers whose number of
         elements is defined at the creation of the container */
         begin
         return false;
         end
      
      proc getIter() : Iter(T)
         // returns an iterator for the container
         begin
         return DListIter.new(first, size);
         end
      
      proc forEach( f : Function(T) )
         // calls "f.exec(x)" for each element x of the container
         begin
         var no : ListNo(T) = first;
         while no <> nil do
            begin
            f.exec(no.getValue());
            no = no.getSucc();
            end
         end

      proc replaceBy( cmd : Command(T) )
            // replaces each element x in the container by "cmd.doIt(x)"
         begin
         var no : ListNo(T) = first;
         while no <> nil do
            begin
            no.replaceBy(cmd);
            no = no.getSucc();
            end
         end

      
      proc collect( f : Filter(T) ) : Container(T)
         /* collects all container elements x such that "f.test(x)"
            evaluates to true. These elements are inserted in a new
            container that is returned by this method. */
            var dlist : DList(T);
         begin
         dlist#init();
         var no : ListNo(T) = first;
         while no <> nil do
            begin
            if f.test(no.getValue())
            then
               dlist.add(no.getValue());
            endif
            no = no.getSucc();
            end
         return dlist;
         end
         
         
      proc reset()
         begin
         current = first;
         end
  
      proc next() : T
         begin
         if current == nil
         then
            return nil;
         else
            var x : T = current.getValue();
            current = current.getSucc();
            return x;
         endif
         end
      
      proc endIter()
      begin
      end

   private:
   
      var first, current : ListNo;
          size : integer;
      
end

        
