

class InputStream subclassOf BasicStream
  proc init()
  
  public:

    proc read( v : array(char)[]; n : long )
             ( exception : CatchFileException )
      // In error, calls raise(ReadFileException)
         
    proc read( v : array(byte)[]; n : long )
             ( exception : CatchFileException )
         
    proc read( s : DynString )
             ( exception : CatchFileException )
         
    proc readln( s : DynString )
               ( exception : CatchFileException )
end


class OutputStream subclassOf BasicStream
  proc init()

  public:
  
    proc write( v : array(char)[] )
              ( exception : CatchFileException )
      // In error, calls raise(WriteFileException) or
      // raise(WriteFileInterException)
              
    proc write( v : array(char)[]; n : long )
              ( exception : CatchFileException )

    proc write( v : array(byte)[] )
              ( exception : CatchFileException )
              
    proc write( v : array(byte)[]; n : long )
              ( exception : CatchFileException )
         
    proc write( s : DynString )
              ( exception : CatchFileException )
         
    proc writeln( s : DynString )
                ( exception : CatchFileException )
end



class Stream subclassOf BasicStream
  proc init()
  
  public:
    // all methods of InputStream and OutputStream
    ...
end


class OpenFileException
  proc init( f : File )
  public:
    proc getFile() : File
end

// there is a class equal to the above for every every file exception.
// There is a throw method in class CatchFileException for each file
// exception

object CatchFileException 
  proc init()
    begin
    singleInstance = CatchFileException.new(' ');
    end
    
  public:
    proc new() : CatchFileException
      begin
      return singleInstance;
      end
end


class CatchFileException
  proc init( x : char )
    begin
      // fake. Just to cause the creation of a new(char) method in 
      // the class object
    end
    
  public:
    proc throw( exc : InvalidNameException )
    proc throw( exc : NonExistingFileException )
      // file does not exist
    proc throw( exc : OpenFileException )
      // file cannot be openend
    proc throw( exc : OpenReadOnlyFileException )
      // attempt to open for writing a read-only file
    proc throw( exc : ReadFileException )
      // attempt to read more bytes than the available or another 
      // read error
    proc throw( exc : WriteSharedFileException )
      // two processes are trying to write to the same file
    proc throw( exc : WriteFileException )
      // no space for writing or other write error
    proc throw( exc : CloseFileException )
      // cannot close file
    proc throw( exc : FileIsClosedException )
      // attempt to use a non-open file
    
    proc raise( exc : InvalidNameException )
              ( exception : CatchFileException ) : String
    proc raise( exc : NonExistingFileException )
              ( exception : CatchFileException ) : String
    proc raise( exc : OpenFileException )
              ( exception : CatchFileException ) : String
    proc raise( exc : OpenReadOnlyFileException )
              ( exception : CatchFileException ) : String
    proc raise( exc : ReadFileException )
              ( exception : CatchFileException )
    proc raise( exc : WriteSharedFileException )
              ( exception : CatchFileException )
    proc raise( exc : WriteFileException )
              ( exception : CatchFileException )
    proc raise( exc : CloseFileException )
              ( exception : CatchFileException )
    proc raise( exc : FileIsClosedException )
              ( exception : CatchFileException )
end    
