class ClassInfo

   proc init( 
         name : String; 
         superclass : ClassInfo; 
         arrayIVI : array(ClassInstanceVariableInfo)[];
         arrayMI : array(ClassMethodInfo)[];
         arrayCI : array(ClassInitMethodInfo)[];
         classObject : AnyClassObject;
         _isAbstract : boolean;
         isReflective : boolean;
         codeAnnotation  : array(CodeAnnotation)[]
         )
      /* arrayIVI is an array with class instance variables. No inherited ones.
         arrayMI is an array with the class methods, no inherited
      */
      begin
      self.name = name;
      self.superclass = superclass;
      self.arrayIVI = arrayIVI;
      self.arrayMI = arrayMI;
      self.arrayCI = arrayCI;
      self.classObject = classObject;
      self._isAbstract = _isAbstract;
      self.isReflective = isReflective;
      self.codeAnnotation = codeAnnotation;
      end
   public:
   
      proc getName() : String
         begin
         return name;
         end
         
      proc toString() : String
         begin
         return name;
         end
         
      proc isSupertypeOf( aSubtype : ClassInfo ) : boolean
            var 
               aSubtypeMethods, selfMethods : array(ClassMethodInfo)[];
               name : String;
               i, j, size : integer;
               found : boolean;
         begin
         aSubtypeMethods = aSubtype.getPublicMethods();
         selfMethods = getPublicMethods();
         if selfMethods.getSize() > aSubtype.getMethods().getSize()
         then
            return false;
         else 
            size = aSubtypeMethods.getSize();
            for i = 0 to selfMethods.getSize() - 1 do
               begin
               j = 0;
               found = false;
               while j < size and not found do 
                  begin
                  if selfMethods[i].matchInterface(aSubtypeMethods[j])
                  then
                     found = true;
                  endif
                  ++j;
                  end
               if not found
               then
                  return false;
               endif
               end
            return true;
         endif
         end
         
      proc isSuperclassOf( aSubclass : ClassInfo ) : boolean
            var aClass : ClassInfo;
         begin
         aClass = aSubclass;
         while aClass <> Any.getAssociateClassInfo() and
               aClass <> AnyValue.getAssociateClassInfo() and
               aClass <> self do
            aClass = aClass.getSuperclass();
         return aClass == self;
         end
         
      proc getSuperclass() : ClassInfo
         begin
         return superclass;
         end
         
      proc getInstanceVariables() : array(ClassInstanceVariableInfo)[]
         begin
         return arrayIVI;
         end
         
      proc getMethods() : array(ClassMethodInfo)[]
            var v : Vector(ClassMethodInfo);
                aClass : ClassInfo;
                anArray : array(ClassMethodInfo)[];
         begin
         v#init();
         aClass = self;
         while aClass <> nil do
            begin
            anArray = aClass.getThisClassMethods();
            for i : integer = 0 to anArray.getSize() - 1 do
               try(HCatchAll)
                  v.add( anArray[i] );
               end
            aClass = aClass.getSuperclass();
            end
         return v.toArray();
         end
         
      proc getPublicMethods() : array(ClassMethodInfo)[]
            var v : Vector(ClassMethodInfo);
                aClass : ClassInfo;
                anArray : array(ClassMethodInfo)[];
         begin
         v#init();
         aClass = self;
         while aClass <> nil do
            begin
            anArray = aClass.getThisClassMethods();
            for i : integer = 0 to anArray.getSize() - 1 do
               begin
               if anArray[i].getVisibility() == MethodInfo.public_v
               then
                  try(HCatchAll)
                     v.add( anArray[i] );
                  end
               endif
               end
            aClass = aClass.getSuperclass();
            end
         return v.toArray();
         end
         
      proc getInitMethods() : array(ClassInitMethodInfo)[]
         begin
         return arrayCI;
         end

         
      proc getInstanceVariable(name : String) : ClassInstanceVariableInfo
            var aClass : ClassInfo;
                anArray : array(ClassInstanceVariableInfo)[];
         begin
         aClass = self;
         while aClass <> nil do
            begin
            anArray = aClass.getThisClassInstanceVariables();
            for i : integer = 0 to anArray.getSize() - 1 do
               if anArray[i].getName().compareTo(name) == 0
               then
                  return anArray[i];
               endif
            aClass = aClass.getSuperclass();
            end
         return nil;
         end

        
      proc getMethod( name : String; paramTypes : array(ClassInfo)[] ) 
             : array(ClassMethodInfo)[]
            var v : Vector(ClassMethodInfo);
                aClass : ClassInfo;
                anArray : array(ClassMethodInfo)[];
         begin
         v#init();
         aClass = self;
         while aClass <> nil do
            begin
            anArray = aClass.getThisClassMethods();
            for i : integer = 0 to anArray.getSize() - 1 do
               if anArray[i].matchInterface(name, paramTypes)
               then
                  try(HCatchAll)
                     v.add( anArray[i] );
                  end
               endif
            aClass = aClass.getSuperclass();
            end
         return v.toArray();
         end
      
      //c
      /* proc getMethod_v( name : String; paramTypes : ... array(AnyClassObject)[] ) : ClassMethodInfo
            var v : Vector(ClassMethodInfo);
                aClass : ClassInfo;
                anArray : array(ClassMethodInfo)[];
         begin
         
         v#init();
         aClass = self;
         while aClass <> nil do
            begin
            anArray = aClass.getThisClassMethods();
            for i : integer = 0 to anArray.getSize() - 1 do
               if anArray[i].matchInterface(name, paramTypes)
               then
                  try(HCatchAll)
                     v.add( anArray[i] );
                  end
               endif
            aClass = aClass.getSuperclass();
            end
         return v.toArray();
         end
         
         */

      proc getPublicMethod( name : String ) : array(ClassMethodInfo)[]
            var v : Vector(ClassMethodInfo);
                aClass : ClassInfo;
                anArray : array(ClassMethodInfo)[];
         begin
         v#init();
         aClass = self;
         while aClass <> nil do
            begin
            anArray = aClass.getThisClassMethods();
            for i : integer = 0 to anArray.getSize() - 1 do
               if anArray[i].getVisibility() == MethodInfo.public_v and
                  anArray[i].getName().compareTo(name) == 0
               then
                  try(HCatchAll)
                     v.add( anArray[i] );
                  end
               endif
            aClass = aClass.getSuperclass();
            end
         return v.toArray();
         end
      
      
      
      proc getThisClassInstanceVariables() : array(ClassInstanceVariableInfo)[]
         begin
         return arrayIVI;
         end
         
      proc getThisClassMethods() : array(ClassMethodInfo)[]
         begin
         return arrayMI;
         end
         
      proc getThisClassPublicMethods() : array(ClassMethodInfo)[]
            var v : Vector(ClassMethodInfo);
                aClass : ClassInfo;
                anArray : array(ClassMethodInfo)[];
         begin
         v#init();
         for i : integer = 0 to arrayMI.getSize() - 1 do
            if arrayMI[i].getVisibility() == MethodInfo.public_v
            then
               try(HCatchAll)
                  v.add( arrayMI[i] );
               end
            endif
         return v.toArray();
         end
         
      proc getAssociateClassObject() : AnyClassObject
         begin
         return classObject;
         end
      
      proc isClassOf( any : AnyClass ) : boolean
         begin
         return self == any.getClassInfo();
         end
         
      proc isAbstract() : boolean
         begin
         return _isAbstract;
         end
         
      proc getNotes() : array(CodeAnnotation)[]
         begin
         return codeAnnotation;
         end
      
   private:
   
      var
         name : String; 
         superclass : ClassInfo; 
         arrayIVI : array(ClassInstanceVariableInfo)[];
         arrayMI : array(ClassMethodInfo)[];
         arrayCI : array(ClassInitMethodInfo)[];
         classObject : AnyClassObject;
         _isAbstract : boolean;
         isReflective : boolean;
         codeAnnotation  : array(CodeAnnotation)[];
         
end