package Interface;

import java.util.*;
import AST.*;
import CompilerError.*;
import java.io.*;
import GreenCompiler.*;

public class ClassTypeInterface extends TypeInterface {
    
    public ClassTypeInterface() {
       super();
       superclass = null;
       superclassName = Type.anyClassType;
       classObjectInterface = null;
       isAbstract = false;
       isReflective = false;
       isShellClass = false;
       hasInterceptAll = false;
       initMethodList = new MethodInterfaceList();
       subclassMethodList = new MethodInterfaceList();
       instanceVariableList = new InstanceVariableList();
    }
    
    
   public void genI( PrintWriter out ) {
    
      if ( classObjectInterface != null ) 
         classObjectInterface.genI(out);
      out.println();
      out.println();
      if ( isReflective )
         out.print("reflective ");
      if ( isAbstract ) 
         out.print("abstract ");
      out.print("class " + name);
      if ( superclassName != null ) {
         out.print( " subclassOf " + superclassName.getGreenName());
      }
      out.println();
      out.println();
      if ( initMethodList.getSize() != 0 ) {
         initMethodList.genI(out);
         out.println();
      }
         
      if ( publicMethodList.getSize() != 0 ) {
         out.println("   public:");
         publicMethodList.genI(out);
         out.println();
      }
      
      if ( subclassMethodList.getSize() != 0 ) {
         out.println("   subclass:");
         subclassMethodList.genI(out);
         out.println();
      }
      
      if ( privateMethodList.getSize() != 0 || instanceVariableList.getSize() != 0 ) {
         out.println("   private:");
         if ( privateMethodList.getSize() != 0 )
            privateMethodList.genI(out);
         if ( instanceVariableList.getSize() != 0 )
            instanceVariableList.genI(out);
         out.println();
      }
      out.println("end");
   }

    public boolean isSupertypeOf( TypeInterface other ) {
        
       if ( ! (other instanceof ClassTypeInterface) ) 
          return super.isSupertypeOf(other);
       else {
          ClassTypeInterface sub = (ClassTypeInterface ) other;
          /*
          while ( sub != null ) 
             if ( sub.getName().compareTo(getName()) == 0 )
                return true;
             else
                sub = sub.getSuperclass();
          */
          if ( ! isSuperclassOf(sub) )
             return super.isSupertypeOf(other);
          else
             return true;
       }
    }
    
    public boolean isSuperclassOf( ClassTypeInterface sub ) {
       while ( sub != null ) 
          if ( sub.getName().compareTo(getName()) == 0 )
             return true;
          else
             sub = sub.getSuperclass();
       return false;
    }
    
    public void reify( TypeInterfaceList til, ErrorMsg em ) {
        
        super.reify(til, em);
        if ( attachedType != null ) 
           attachedType.reify(til, em);
           
        if ( superclassName != null ) {
           superclass = til.get(superclassName.getInterfaceName());
           if ( superclass == null ) {
              em.show("Class " + superclassName.getInterfaceName() + " does not exist. It is used as a superclass of " + name);
              return;
           }
        }
        
        if ( subclassMethodList != null ) 
           subclassMethodList.reify(til, em);
        if ( initMethodList != null )
           initMethodList.reify(til, em);
        if ( classObjectInterface != null ) 
           classObjectInterface.reify(til, em);
        if ( instanceVariableList != null )
           instanceVariableList.reify(til, em);
    
           
    }
    

    public MethodInterface publicSearch( String methodSearchName ) {
       // search for method using its searchName. Superclasses, if they exist, are used
       
       ClassTypeInterface current = this;
       
       while ( current != null ) {
          MethodInterface m = current.getPublicMethodList().get(methodSearchName);
          if ( m != null ) 
             return m;
          current = current.getSuperclass();
       }
       return null;
    }
    
    public Hashtable multiplePublicSearch( String anySearchMethodName, Hashtable result ) {
       
       ClassTypeInterface current = this;
       
       while ( current != null ) {
          current.getPublicMethodList().multipleGet(anySearchMethodName, result);
          current = current.getSuperclass();
       }
             
       return result;
    }
       
    
    public MethodInterface simpleSubclassSearch( String methodSearchName ) {
       // search for method using its searchName. The search is made only in this
       // type. The superclass, it this is a class, is not searched for
       return (MethodInterface ) subclassMethodList.get(methodSearchName);
    }

    public MethodInterface simplePrivateSearch( String methodSearchName ) {
       // search for method using its searchName. The search is made only in this
       // type. The superclass, it this is a class, is not searched for
       return (MethodInterface ) privateMethodList.get(methodSearchName);
    }

    public MethodInterface searchSubclass( String methodSearchName ) {
       // search for method using its searchName
       
       ClassTypeInterface current = this;
       
       while ( current != null ) {
          MethodInterface m = current.simpleSubclassSearch(methodSearchName);
          if ( m != null ) 
             return m;
          current = current.getSuperclass();
       }
       return null;
    }
    
    public ClassTypeInterface getSuperclass() {
        return superclass;
    }
    
    public ClassType getSuperclassName() {
        return superclassName;
    }
    
    public void setSuperclassName( ClassType superclassName ) {
        this.superclassName = superclassName;
    }
    
    public ClassObjectTypeInterface getClassObjectInterface() {
        return classObjectInterface;
    }
    
    public void setClassObjectInterface( ClassObjectTypeInterface classObjectInterface ) {
        this.classObjectInterface = classObjectInterface;
    }
    
    public boolean getIsAbstract() {
        return isAbstract;
    }
    
    public void setIsAbstract( boolean isAbstract ) {
        this.isAbstract = isAbstract;
    }
    
    public boolean getIsReflective() {
        return isReflective;
    }
    
    public void setIsReflective( boolean isReflective ) {
        this.isReflective = isReflective;
    }

    public void setInstanceVariableList( InstanceVariableList instanceVariableList ) {
        this.instanceVariableList = instanceVariableList;
    }
    
    public InstanceVariableList getInstanceVariableList() {
        return instanceVariableList;
    }

    public void setInitMethodList( MethodInterfaceList initMethodList ) {
        this.initMethodList = initMethodList;
    }
    
    public MethodInterfaceList getInitMethodList() {
        return initMethodList;
    }
    
    public void setSubclassMethodList( MethodInterfaceList subclassMethodList ) {
        this.subclassMethodList = subclassMethodList;
    }
    
    public MethodInterfaceList getSubclassMethodList() {
        return subclassMethodList;
    }
    
    public ClassFile getClassFile() {
        return classFile;
    }
    
    public void setClassFile( ClassFile classFile ) {
        this.classFile = classFile;
    }
    
    public MethodInterface subclassPublicSearch( String methodSearchName ) {
    
       ClassTypeInterface current = this;
       while ( current != null ) {
          MethodInterface mi = current.getPublicMethodList().get(methodSearchName);
          if ( mi != null )
             return mi;
          mi = current.getSubclassMethodList().get(methodSearchName);
          if ( mi != null )
             return mi;
          if ( (mi = current.getInitMethodList().get(methodSearchName)) != null )
             return mi;
          current = current.getSuperclass();
       }
       return null;
    }
    
    public Hashtable multipleSubclassPublicSearch(String anySearchMethodaName) {
        
        Hashtable result = new Hashtable();
        
        ClassTypeInterface current = this;
        while ( current != null ) {
            current.getPublicMethodList().multipleGet(anySearchMethodaName, result);
            current.getSubclassMethodList().multipleGet(anySearchMethodaName, result);
            current.getInitMethodList().multipleGet(anySearchMethodaName, result);
            current = current.getSuperclass();
        }
        return result;
    }
             
    public MethodInterface fullSearch( String methodSearchName ) {
        ClassTypeInterface current = this;
        MethodInterface mi;
        
        if ( (mi = current.getPrivateMethodList().get(methodSearchName)) != null ||
             (mi = current.getPublicMethodList().get(methodSearchName)) != null ||
             (mi = current.getSubclassMethodList().get(methodSearchName)) != null ||
             (mi = current.getInitMethodList().get(methodSearchName)) != null )
             return mi;
      
        current = current.getSuperclass();
        while ( current != null ) {
           if ( (mi = current.getPublicMethodList().get(methodSearchName)) != null ||
                (mi = current.getSubclassMethodList().get(methodSearchName)) != null ||
                (mi = current.getInitMethodList().get(methodSearchName)) != null )
              return mi;
           current = current.getSuperclass();
        }
        return null;
    }

    public MethodInterface fullWrapperSearch( 
                         String methodName,
                         String methodSearchWrapperName ) {
        ClassTypeInterface current = this;
        MethodInterface mi;
        
        if ( (mi = current.getPrivateMethodList().getWrapper(methodName, methodSearchWrapperName)) != null ||
             (mi = current.getPublicMethodList().getWrapper(methodName, methodSearchWrapperName)) != null ||
             (mi = current.getSubclassMethodList().getWrapper(methodName, methodSearchWrapperName)) != null ||
             (mi = current.getInitMethodList().getWrapper(methodName, methodSearchWrapperName)) != null )
             return mi;
      
        current = current.getSuperclass();
        while ( current != null ) {
           if ( (mi = current.getPublicMethodList().getWrapper(methodName, methodSearchWrapperName)) != null ||
                (mi = current.getSubclassMethodList().getWrapper(methodName, methodSearchWrapperName)) != null ||
                (mi = current.getInitMethodList().getWrapper(methodName, methodSearchWrapperName)) != null )
              return mi;
           current = current.getSuperclass();
        }
        return null;
    }                            
    

    public Hashtable multipleFullSearch( String methodName, ExprList exprList ) {
        return multipleFullSearch( NameServer.buildAnySearchName(methodName, exprList) );
    }
    
    public Hashtable multipleFullSearch( String anySearchMethodName ) {
        
        Hashtable result = new Hashtable();
        
        ClassTypeInterface current = this;
        
        current.getPrivateMethodList().multipleGet(anySearchMethodName, result);
        current.getPublicMethodList().multipleGet(anySearchMethodName, result);
        current.getSubclassMethodList().multipleGet(anySearchMethodName, result);
        current.getInitMethodList().multipleGet(anySearchMethodName, result);
             
        current = current.getSuperclass();
        while ( current != null ) {
           current.getPublicMethodList().multipleGet(anySearchMethodName, result);
           current.getSubclassMethodList().multipleGet(anySearchMethodName, result);
           current.getInitMethodList().multipleGet(anySearchMethodName, result);
           current = current.getSuperclass();
        }
        return result;
    }
    
    public String getJavaFileName() {
       if ( javaFileName == null ) {
          if ( ! isShellClass ) 
             javaFileName = NameServer.getJavaFileName(name, getJavaName());
          else
             javaFileName = getJavaName();
       }
       return javaFileName;
    }
    
    public String getJavaName() {
       if ( javaName == null ) {
          if ( ! isShellClass ) 
             javaName = NameServer.buildJavaName(name);
          else
             javaName = NameServer.getShellClassName(name, attachedType);
       }
       return javaName;
    }
       
       
    public String getInterfaceName() {
       if ( NameServer.basicTypes.get(getName()) == null )
          return getJavaName();
       else
          return "_" + getName();
    }
    
    public String getAccessToClassInfo() {
       return NameServer.getAccessToClassInfoClass( 
           getIsBasicType() || getName().compareTo("String") == 0 ? 
                NameServer.buildBasicTypeClassName(getName()) : getJavaName() );
    }
    
       
    public void genInterfaceNameList( PrintWriter out ) {
       
       boolean first = true;
       Hashtable h = new Hashtable();
       ClassTypeInterface current = this;
       while ( current != null ) {
          MethodInterfaceList mil = current.getPublicMethodList();
          Enumeration e = mil.anyElements();
          while ( e.hasMoreElements() ) {
             String interfaceName = ((MethodInterface ) e.nextElement()).getMethodInterfaceName();
             if ( h.put(interfaceName, "") == null ) {
                if ( ! first )
                   out.println(",");
                ws.print( interfaceName );
                first = false;
             }
          }
          current = current.getSuperclass();
       }
     }

    public boolean getIsShellClass() {
       return isShellClass;
    }
    
    public void setIsShellClass( boolean isShellClass ) {
       this.isShellClass = isShellClass;
    }
    
    public ClassType getAttachedType() {
       return attachedType;
    }
    
    public void setAttachedType( ClassType attachedType ) {
       this.attachedType = attachedType;
    }

    public String getShell_ivcName() {
       if ( shell_ivcName == null ) { 
          shell_ivcName = NameServer.getShell_ivClassName(getJavaName());
       }
       return shell_ivcName;
    }
       
    public void setHasInterceptAll( boolean hasInterceptAll ) {
       this.hasInterceptAll = hasInterceptAll;
    }
    
    public boolean getHasInterceptAll() {
       return hasInterceptAll;
    }
    
    
    private String shell_ivcName;
    private ClassFile classFile;
    private String javaName, javaFileName;
    
    protected MethodInterfaceList subclassMethodList;
    protected MethodInterfaceList initMethodList;
    
    protected ClassTypeInterface superclass;
    protected ClassType superclassName;
    protected ClassType attachedType;  // for shell classes
    protected ClassObjectTypeInterface classObjectInterface;
       // if this represents a class, is it abstract ?
    protected boolean isAbstract;
    protected boolean isReflective;
    protected InstanceVariableList instanceVariableList;
    protected boolean isShellClass;
    protected boolean hasInterceptAll;
    
}
