package Interface;

import CompilerError.*;
import java.util.*;
import java.io.*;
import AST.*;
import GreenCompiler.*;

/*
   TypeInterface may represent a class or class object. 
   If this interface represents a class object, classObjectInterface is null.
*/

abstract public class TypeInterface {

    public TypeInterface() {
        name = null;
        privateMethodList = new MethodInterfaceList();
        publicMethodList = new MethodInterfaceList();
        isBasicType = false;
        isJavaClass = false;
        hasThrowMethod = false;
        hasRaiseMethod = false;
    }

    
    abstract public void genI( PrintWriter out );
    
    public void reify( TypeInterfaceList til, ErrorMsg errorMsg ) {

       if ( privateMethodList != null )
          privateMethodList.reify(til, errorMsg);
       if ( publicMethodList != null )
          publicMethodList.reify(til, errorMsg);
        
    }
    

    public boolean isSupertypeOf( TypeInterface other ) {
       
       if ( other.getName().compareTo("Nil") == 0 ||
            name.compareTo(other.getName()) == 0 )
            return true;
       Enumeration e = publicMethodList.elements();
       while ( e.hasMoreElements() ) {
          MethodInterface m = publicMethodList.get( (String ) e.nextElement());
          if ( other.publicSearch( m.getSearchName() ) == null ) 
             return false;
       }
       return true;
    }


    public MethodInterfaceList subtractTypeFrom( TypeInterface other ) {
        // return a Vector of MethodInterface
       MethodInterfaceList result = new MethodInterfaceList();
       
       if ( other.getName().compareTo("Nil") == 0 ||
            name.compareTo(other.getName()) == 0 )
            return result;
       Enumeration e = publicMethodList.elements();
       while ( e.hasMoreElements() ) {
          MethodInterface m = publicMethodList.get( (String ) e.nextElement());
          if ( other.publicSearch( m.getSearchName() ) == null ) 
             result.add(m);
       }
       return result;
    }


    abstract public void genInterfaceNameList( PrintWriter out );
    
    
    public MethodInterface publicSearch( String methodName, ExprList exprList ) {
       // typeArray has types with searchNames
       
       return publicSearch(NameServer.buildMethodSearchName(methodName, exprList));
    }
    

    
    public MethodInterface publicSearch( String methodSearchName ) {
       return publicMethodList.get(methodSearchName);
    }
    

    public Hashtable multiplePublicSearch( String methodName, ExprList exprList ) {
        return multiplePublicSearch( NameServer.buildAnySearchName(methodName, exprList) ); 
    }
       
    public Hashtable multiplePublicSearch( String anySearchMethodName ) {
       /* return a Vector with all methods of the interface with name anySearchMethodName.
          */
       return multiplePublicSearch(anySearchMethodName, new Hashtable());
    }
       
    public Hashtable multiplePublicSearch( String anySearchMethodName, Hashtable result ) {
       return publicMethodList.multipleGet(anySearchMethodName, result);
    }
        
          
    public String getName() {
        // return the name of the class or class object of the interface. 
        // Probably this method will not be used
        return name;
    }
    
    public void setName( String name ) {
        this.name = name;
    }
    
    public void setPublicMethodList( MethodInterfaceList publicMethodList ) {
        this.publicMethodList = publicMethodList;
    }
    
    public MethodInterfaceList getPublicMethodList() {
        return publicMethodList;
    }
    
    public void setPrivateMethodList( MethodInterfaceList privateMethodList ) {
        this.privateMethodList = privateMethodList;
    }
    
    public MethodInterfaceList getPrivateMethodList() {
        return privateMethodList;
    }
    
    public Constant searchConstant( String name ) {
        return null;
    }
    
    
    public MethodInterface fullSearch( String methodName, ExprList exprList ) {
        return fullSearch( NameServer.buildMethodSearchName(methodName, exprList) );
    }
    
    
    public Hashtable multipleFullSearch( String methodName, ExprList exprList ) {
        return multipleFullSearch( NameServer.buildAnySearchName(methodName, exprList) );
    }
    
    abstract public Hashtable multipleFullSearch( String anySearchMethodName );
    
    
    public void setIsJavaClass( boolean isJavaClass ) {
       this.isJavaClass = isJavaClass;
    }
    
    public boolean getIsJavaClass() {
       return isJavaClass;
    }
    
    abstract public MethodInterface fullSearch(String methodSearchName);
    
    abstract public MethodInterface fullWrapperSearch( 
                         String methodName,
                         String methodSearchWrapperName );
    
    
    public Constant fullSearchConstant( String name ) {
        return null;
    }
    
    
    abstract public String getJavaName();
    
    public String getInterfaceName() {
       return getJavaName();
    }

    
    public void setToAddTextPublic( char []publicText ) {
        this.publicText = publicText;
    }
    
    public char []getToAddTextPublic() {
        return publicText;
    }

    public void setToAddTextPrivate( char []privateText ) {
        this.privateText = privateText;
    }
    
    public char []getToAddTextPrivate() {
        return privateText;
    }
    
    abstract public String getAccessToClassInfo();

    public void setIsBasicType( boolean isBasicType ) {
       this.isBasicType = isBasicType;
    }
    
    public boolean getIsBasicType() {
       return isBasicType;
    }
    
    public void setHasThrowMethod( boolean hasThrowMethod ) {
       this.hasThrowMethod = hasThrowMethod;
    }
    
    public boolean getHasThrowMethod() { 
       return hasThrowMethod;
    }
    
    public void setHasRaiseMethod( boolean hasRaiseMethod ) {
       this.hasRaiseMethod = hasRaiseMethod;
    }
    
    public boolean getHasRaiseMethod() { 
       return hasRaiseMethod;
    }
    
    
    
    
    private boolean isBasicType;
    private char []publicText;
    private char []privateText;
    
    protected MethodInterfaceList privateMethodList;
    protected String name;
    protected String javaName;
      // public methods
    protected MethodInterfaceList publicMethodList;
    
    protected boolean isJavaClass;
    protected boolean hasThrowMethod;
    protected boolean hasRaiseMethod;
    
}
    


    